import 'dart:async';

import 'package:build/build.dart';
import 'package:cactus_sync_client_gen/src/gql_dart_formatter.dart';
import 'package:cactus_sync_client_gen/src/gql_enum.dart';
import 'package:cactus_sync_client_gen/src/gql_input_converter.dart';
import 'package:cactus_sync_client_gen/src/gql_model_builder.dart';
import 'package:cactus_sync_client_gen/src/used_directives.dart';
import 'package:code_builder/code_builder.dart';
import 'package:gql/language.dart' as gql_lang;
import "package:gql/schema.dart" as gql_schema;
import 'package:indent/indent.dart';

class ModelBuilder implements Builder {
  @override
  FutureOr<void> build(BuildStep buildStep) async {
    // Retrieve the currently matched asset
    final inputId = buildStep.inputId;
    final assetName = inputId.pathSegments.last.split('.').first;

    /// Create new target
    final copyAssetId = inputId.changeExtension('.dart');
    final originContentStr = await buildStep.readAsString(inputId);
    final schemaDocument = gql_lang.parseString(originContentStr);
    final schema = gql_schema.buildSchema(schemaDocument);
    // TODO: add operation types
    // TODO: add enums

    // TODO: add scalars
    Library library = Library();
    void rebuildLibrary(Iterable<Spec> iterable) =>
        library = library.rebuild((l) => l.body.addAll(iterable));
    final operationTypes = schema.typeMap;
    final modelsAndProviders = GqlModelBuilder().makeModelsAndProviders(
      operationTypes: operationTypes.values,
    );
    rebuildLibrary(modelsAndProviders.models);

    final inputClasses = GqlInputs().fromSchema(
      inputObjectTypes: schema.inputObjectTypes,
    );
    rebuildLibrary(inputClasses);

    final dartEnums = GqlEnums.fromSchema(
      schemaEnums: schema.enums,
    );
    rebuildLibrary(dartEnums);

    final directivesStr = [
      ...usedDirectives.map(
        (e) => Directive.import(e),
      ),
      Directive.part('$assetName.g.dart'),
    ]
        .map(
          (d) => GqlDartFormatter.formatAndStringify(spec: d),
        )
        .join("\n");

    final finalBuffer = StringBuffer(
      """
        /// !------------ CAUTION ------------!
        /// Autogenerated file. Please do not edit it manually!
        /// Updated: ${DateTime.now()}
        /// !---------- END CAUTION ----------!
        
        import 'package:graphql/client.dart' as graphql;
        $directivesStr
      """
          .unindent(),
    );

    final libraryStr = GqlDartFormatter.formatAndStringify(spec: library);
    finalBuffer..writeln(libraryStr)..writeln(modelsAndProviders.providers);
    final finalContent = finalBuffer.toString();

    await buildStep.writeAsString(copyAssetId, finalContent.unindent());
  }

  @override
  Map<String, List<String>> get buildExtensions => {
        '.graphql': [".dart"]
      };
}
